from pydantic import BaseModel, Field

from ..buffered_ml_messages import BufferedMLMessageType
from ..templating import TemplateInputs
from .params import StaticTool
from .settings import TCallModelSettings
from .specs import ToolJsonSchema, ToolSpec


class RawTCallRequest(BaseModel):
    tools: list[ToolSpec] | list[ToolJsonSchema] | list[str] = Field(default_factory=list)
    static_tools: list[StaticTool] = Field(default_factory=list)

    messages: list[BufferedMLMessageType]

    settings: TCallModelSettings


class PromptedTCallRequest(BaseModel):
    tools: list[ToolSpec] | list[ToolJsonSchema] | list[str] = Field(default_factory=list)
    static_tools: list[StaticTool] = Field(default_factory=list)

    prompt_inputs: list[TemplateInputs]
    prompt_name: str
    settings: TCallModelSettings | None = None


class HybridTCallRequest(BaseModel):
    tools: list[ToolSpec] | list[ToolJsonSchema] | list[str] = Field(default_factory=list)
    static_tools: list[StaticTool] = Field(default_factory=list)

    messages: list[BufferedMLMessageType]

    prompt_inputs: list[TemplateInputs]
    prompt_name: str

    settings: TCallModelSettings | None = None



from typing import Any

from pydantic import BaseModel, Field


class StaticParams(BaseModel):
    query: dict[str, Any] = Field(default_factory=dict)
    body: dict[str, Any] = Field(default_factory=dict)


class GeneratedParams(BaseModel):
    path: list[str] = Field(default_factory=list)
    query: list[str] = Field(default_factory=list)
    body: list[str] = Field(default_factory=list)


class Forwardings(BaseModel):
    headers: list[str] = Field(default_factory=list)
    query: list[str] = Field(default_factory=list)
    body: list[str] = Field(default_factory=list)


class StaticTool(BaseModel):
    name: str
    arguments: dict[str, Any]
    response: Any | None = None



from typing import Annotated

from pydantic import BaseModel, Field


class BaseModelSettings(BaseModel):
    model: str
    max_tokens: int | None = Field(default=None, ge=1)  # defaults to inf
    temperature: float | None = Field(default=None, ge=0, le=2)
    top_p: float | None = Field(default=None, ge=0, le=1)
    logit_bias: dict[str, Annotated[int, Field(ge=-100, le=100)]] | None = None
    stop: list[str] | None = Field(default=None, max_length=4)
    presence_penalty: float = Field(default=0, ge=-2, le=2)
    frequency_penalty: float = Field(default=0, ge=-2, le=2)



from ..base_settings import BaseModelSettings


class ChatModelSettings(BaseModelSettings):
    """
    Change these settings to tweak the model's behavior.

    Based on https://platform.openai.com/docs/api-reference/chat/create
    """



from datetime import datetime

from melting_schemas.utils import wrap

from ..buffered_ml_messages import ChatMLMessage
from ..templating import TemplateInputs
from .requests import (
    ChatCompletionRequest,
    HybridChatCompletionRequest,
    RawChatCompletionRequest,
)
from .settings import ChatModelSettings


def raw_chat_completion_request_examples():
    raw = RawChatCompletionRequest(
        messages=[
            ChatMLMessage(content="You are a helpful chatbot.", role="system"),
            ChatMLMessage(content="What does bequeath mean?", role="user"),
        ],
        settings=ChatModelSettings(model="openai/gpt-3.5-turbo-1106"),
    )
    named_raw = RawChatCompletionRequest(
        messages=[
            ChatMLMessage(content="You are a helpful chatbot.", role="system"),
            ChatMLMessage(content="What does my name mean?", name="John", role="user"),
        ],
        settings=ChatModelSettings(model="openai/gpt-4-0613"),
    )

    return [wrap(name="Raw", value=raw), wrap(name="Named Raw", value=named_raw)]


def chat_completion_request_examples():
    prompted = ChatCompletionRequest(
        prompt_inputs=[
            TemplateInputs(role="system", inputs={"now": str(datetime.now())}),
            TemplateInputs(role="user", inputs={"text": "What day is today?"}),
        ],
        prompt_name="teia.example.1",
    )
    many_templates = ChatCompletionRequest(
        prompt_inputs=[
            TemplateInputs(
                role="system",
                inputs={"plugin_data": "Secret number is 42"},
                template_name="plugin_prompt",
            ),
            TemplateInputs(
                role="user",
                inputs={"question": "What is the secret number???"},
                template_name="user_prompt",
            ),
        ],
        prompt_name="teia.example.2",
    )
    return [
        wrap(name="Prompted", value=prompted),
        wrap(name="Many Templates", value=many_templates),
    ]


def hybrid_request_examples():
    hybrid = HybridChatCompletionRequest(
        messages=[
            ChatMLMessage(content="You are a helpful chatbot.", role="system"),
            TemplateInputs(role="user", inputs={"text": "What day is today?"}),
        ],
        prompt_name="teia.example.1",
    )
    return [wrap(name="Hybrid", value=hybrid)]



from ..base_settings import BaseModelSettings


class TextModelSettings(BaseModelSettings):
    """
    Change these settings to tweak the model's behavior.

    Base on https://platform.openai.com/docs/api-reference/completions/create
    """

    n: int = 1



from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, Field, SerializeAsAny


class BaseSchema(BaseModel):
    """
    Generic keywords shared amongst schemas based on JSON Schema spec:
    https://json-schema.org/understanding-json-schema/reference/generic
    """

    enum: list["Schemas"] = Field(default_factory=list)

    # Docs
    comment: str = Field(default="", alias="$comment")
    description: str = ""
    title: str = ""

    # Examples
    default: Any = None
    examples: list[Any] = Field(default_factory=list)

    # Flags
    deprecated: bool = False
    read_only: bool = Field(default=False, alias="readOnly")
    write_only: bool = Field(default=False, alias="writeOnly")

    model_config = ConfigDict(populate_by_name=True, frozen=True)


class StringSchema(BaseSchema):
    """
    String schema following JSON Schema spec:
    https://json-schema.org/understanding-json-schema/reference/numeric
    """

    type: Literal["string"]

    # Built-in formats: https://json-schema.org/understanding-json-schema/reference/string#built-in-formats
    format: str = ""

    # Validation
    pattern: str = ""
    max_length: int = Field(default=0, alias="maxLength")
    min_length: int = Field(default=0, alias="minLength")


class NumberSchema(BaseSchema):
    """
    Boolean schema following JSON Schema spec:
    https://json-schema.org/understanding-json-schema/reference/numeric
    """

    type: Literal["integer", "number"]
    multiple_of: int = Field(default=0, alias="multipleOf")
    minimum: int = Field(default=0, alias="minimum")
    maximum: int = Field(default=0, alias="maximum")

    exclusive_minimum: int = Field(default=0, alias="exclusiveMinimum")
    exclusive_maximum: int = Field(default=0, alias="exclusiveMaximum")


class BooleanSchema(BaseSchema):
    """
    Boolean schema following JSON Schema spec:
    https://json-schema.org/understanding-json-schema/reference/boolean
    """

    type: Literal["boolean"]


class NullSchema(BaseSchema):
    """
    Null schema following JSON Schema spec:
    https://json-schema.org/understanding-json-schema/reference/null
    """

    type: Literal["null"]


class ArraySchema(BaseSchema):
    """
    Array schema following JSON Schema spec:
    https://json-schema.org/understanding-json-schema/reference/array
    """

    type: Literal["array"]
    items: "Schemas" = Field(default_factory=list)
    prefix_items: list[SerializeAsAny["Schemas"]] = Field(default_factory=list, alias="prefixItems")
    additional_items: SerializeAsAny["Schemas"] = Field(
        default_factory=list, alias="additionalItems"
    )
    max_items: int = Field(default=0, alias="maxItems")
    min_items: int = Field(default=0, alias="minItems")
    unique_items: bool = Field(default=False, alias="uniqueItems")
    contains: SerializeAsAny["Schemas | None"] = Field(default=None)
    min_contains: int = Field(default=0, alias="minContains")
    max_contains: int = Field(default=0, alias="maxContains")


class ObjectSchema(BaseSchema):
    """
    Object schema following JSON Schema spec:
    https://json-schema.org/understanding-json-schema/reference/object
    """

    type: Literal["object"]
    max_properties: int = Field(default=0, alias="maxProperties")
    min_properties: int = Field(default=0, alias="minProperties")
    properties: dict[str, SerializeAsAny["Schemas"]] = Field(default_factory=dict)
    required: list[str] = Field(default_factory=list)


PrimitiveSchemas = StringSchema | NumberSchema | BooleanSchema | NullSchema
Schemas = ArraySchema | ObjectSchema | PrimitiveSchemas


class FunctionJsonSchema(BaseModel):
    name: str
    description: str = ""
    parameters: ObjectSchema | None = None


BaseSchema.model_rebuild()
ArraySchema.model_rebuild()
ObjectSchema.model_rebuild()



from typing import Any

from bson import ObjectId
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema


class PyObjectId:
    @classmethod
    def validate_object_id(cls, v: Any, handler) -> ObjectId:
        if isinstance(v, ObjectId):
            return v

        s = handler(v)
        if ObjectId.is_valid(s):
            return ObjectId(s)
        else:
            raise ValueError("Invalid ObjectId")

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, _handler) -> core_schema.CoreSchema:
        assert source_type is ObjectId
        return core_schema.no_info_wrap_validator_function(
            cls.validate_object_id,
            core_schema.str_schema(),
            serialization=core_schema.to_string_ser_schema(),
        )

    @classmethod
    def __get_pydantic_json_schema__(cls, _core_schema, handler) -> JsonSchemaValue:
        return handler(core_schema.str_schema())



from melting_schemas.utils import wrap

from .requests import RawTextCompletionRequest
from .settings import TextModelSettings


def text_completion_examples():
    minimal = RawTextCompletionRequest(
        text="Hello World",
        settings=TextModelSettings(model="openai/gpt-3.5-turbo-instruct"),
    )
    with_suffix = RawTextCompletionRequest(
        text="# write a function that returns the sum of two numbers\n",
        settings=TextModelSettings(model="openai/gpt-3.5-turbo-instruct"),
        suffix="def sum(a, b):\n    return a + b\n",
    )

    return [wrap(name="Minimal", value=minimal), wrap(name="With Suffix", value=with_suffix)]



from pydantic import BaseModel

from ..usage import TokenUsage


class TextEncoding(BaseModel):
    model: str
    snippets: list[str]
    vectors: list[list[float]]

    usage: TokenUsage | None = None


class RawTextEncodingRequest(BaseModel):
    snippets: list[str]
    model: str
    dims: int | None = None



from typing import Any

from prompts import (
    DynamicSchema,
    OpenAIModelSettings,
    PromptRole,
    Template,
    TurboSchema,
)
from prompts.schemas import TurboSchema
from pydantic import BaseModel, Field

from melting_schemas.completion.buffered_ml_messages import ChatMLMessage
from melting_schemas.completion.templating import TemplateInputs

from ..completion.chat import ChatModelSettings

# ====== Create Schemas ======


class ChatPromptTemplate(TurboSchema):
    settings: ChatModelSettings


class CreateCompletionPrompt(DynamicSchema):
    pass


# ====== Get Schemas ======


class GeneratedFields(BaseModel):
    created_at: str
    created_by: Any
    id: str = Field(alias="_id")


class ChatPrompt(GeneratedFields, TurboSchema):
    pass


class GetCompletionPrompt(GeneratedFields, DynamicSchema):
    pass


# ====== Update Schemas ======


class UpdateChatTemplateData(BaseModel):
    name: str | None = None
    role: PromptRole | None = None
    replacements: dict[str, str] | None = None


class UpdateSettings(BaseModel):
    model: str | None = None
    max_tokens: int | None = None
    stop: list[str] | None = None
    temperature: float | None = None
    top_p: float | None = None
    frequency_penalty: float | None = None
    presence_penalty: float | None = None


class BaseUpdatePrompt(BaseModel):
    # Prompt fields
    description: str | None = None
    settings: UpdateSettings | None = None

    # Prompt start
    initial_template_data: str | list[UpdateChatTemplateData] | None = None


class UpdateChatPrompt(BaseUpdatePrompt):
    # Templates
    assistant_templates: list[Template] | None = None
    system_templates: list[Template] | None = None
    user_templates: list[Template] | None = None


class UpdateCompletionPrompt(BaseUpdatePrompt):
    # Template
    template: str | None = None


# ========= Full Schema ==============


class CompleteBasePromptTemplate(BaseModel):
    name: str
    description: str = ""
    settings: ChatModelSettings
    initial_template_data: list[ChatMLMessage | TemplateInputs] = Field(
        default_factory=list
    )


class CompleteChatPromptTemplate(CompleteBasePromptTemplate):
    settings: ChatModelSettings

    # chat templates
    assistant_templates: list[Template] | str
    system_templates: list[Template] | str
    user_templates: list[Template] | str


class CompleteTextPromptTemplate(CompleteBasePromptTemplate):
    settings: OpenAIModelSettings

    # Chat Templates
    assistant_templates: list[Template] | None = None
    system_templates: list[Template] | None = None
    user_templates: list[Template] | None = None

    # Completion Template
    template: str | None = None



from typing import Literal

from pydantic import BaseModel, Field

from melting_schemas.completion.base_settings import BaseModelSettings
from melting_schemas.completion.buffered_ml_messages import (
    BufferedMLMessageType,
    ChatMLMessage,
    ToolCallMLMessage,
)
from melting_schemas.completion.finish_reason import FinishReason
from melting_schemas.completion.tcall import ToolInfo
from melting_schemas.completion.templating import Templating
from melting_schemas.usage import StreamTimings, Timings, TokenUsage


class CompletionResponse[T: (StreamTimings, Timings), S: BaseModelSettings](BaseModel):
    # Context
    messages: list[BufferedMLMessageType] = Field(default_factory=list)
    tools: list[ToolInfo] = Field(default_factory=list)
    templating: Templating | None = None

    # Out
    output: ChatMLMessage | ToolCallMLMessage | str

    # Analytics
    finish_reason: FinishReason
    timings: T
    usage: TokenUsage

    # Config
    type: Literal["chat", "tcall", "text"]
    settings: S



from pydantic import BaseModel

from ..buffered_ml_messages import ChatMLMessage
from ..templating import TemplateInputs
from .settings import ChatModelSettings


class RawChatCompletionRequest(BaseModel):
    messages: list[ChatMLMessage]
    settings: ChatModelSettings


class ChatCompletionRequest(BaseModel):
    prompt_inputs: list[TemplateInputs]
    prompt_name: str
    settings: ChatModelSettings | None = None


class HybridChatCompletionRequest(BaseModel):
    prompt_name: str
    messages: list[ChatMLMessage | TemplateInputs]
    settings: ChatModelSettings | None = None



from melting_schemas.completion.tcall.params import (
    Forwardings,
    GeneratedParams,
    StaticParams,
)
from melting_schemas.completion.tcall.specs import (
    HttpToolCallee,
    ToolJsonSchema,
    ToolSpec,
)
from melting_schemas.json_schema import (
    ArraySchema,
    FunctionJsonSchema,
    ObjectSchema,
    StringSchema,
)
from melting_schemas.utils import wrap


def create_tool_examples():
    tool_with_static_call = ToolSpec(
        callee=HttpToolCallee(
            type="http",
            method="POST",
            forwardings=Forwardings(headers=[]),
            headers={"authorization": "Bearer <melt-token>"},
            url="https://search.allai.digital/search/",
            static_params=StaticParams(query={}, body={}),
            generated_params=GeneratedParams(path=[], query=[], body=["query", "search_settings"]),
        ),
        json_schema=ToolJsonSchema(
            type="function",
            function=FunctionJsonSchema(
                name="rag-search-salesforce-help",
                description="Perform a semantic search on Salesforce Help documentation pages (https://help.salesforce.com/).",
                parameters=ObjectSchema(
                    type="object",
                    properties={
                        "query": StringSchema(
                            type="string",
                            description="The search query to be embedded.",
                        ),
                        "search_settings": ArraySchema(
                            type="array",
                            items=ObjectSchema(
                                type="object",
                                properties={
                                    "kb_name": StringSchema(
                                        type="string",
                                        description="Enum: ['salesforce_help_may_2024'].",
                                    )
                                },
                                required=["kb_name"],
                            ),
                        ),
                    },
                    required=["query", "search_settings"],
                ),
            ),
        ),
        tool_name="temp-help-salesforce-search",
    )

    return [wrap(name="Basic Tool Call", value=tool_with_static_call)]



from typing import Literal

from ..base_settings import BaseModelSettings


class TCallModelSettings(BaseModelSettings):
    """
    Change these settings to tweak the model's behavior.

    Base on https://platform.openai.com/docs/api-reference/chat/create
    """

    max_iterations: int = 10  # Maximum back and fourth allowed
    tool_choice: Literal["auto", "required"] = "auto"



from ..buffered_ml_messages import ChatMLMessage
from ..streamed_ml_messages import ChatMLMessageChunk
from ..templating import TemplateInputs, Templating
from .examples import (
    chat_completion_request_examples,
    hybrid_request_examples,
    raw_chat_completion_request_examples,
)
from .requests import (
    ChatCompletionRequest,
    HybridChatCompletionRequest,
    RawChatCompletionRequest,
)
from .settings import ChatModelSettings



from typing import Protocol

from fastapi.openapi.models import Example
from pydantic import BaseModel


class NamedDumpable(Protocol):
    name: str

    model_dump = BaseModel.model_dump


class OpenAPIExample(BaseModel):
    name: str
    value: BaseModel


def wrap(name: str, value: BaseModel) -> NamedDumpable:
    return OpenAPIExample(name=name, value=value)


def to_openapi_examples(examples: list[NamedDumpable | OpenAPIExample]) -> dict[str, Example]:
    """
    Convert a list of objects to a dictionary of OpenAPI Examples.

    It accepts a list of NamedDumpable compatible objects or OpenAPIExample objects.

    NamedDumpable objects are expected to have a `name` attribute and a `model_dump` method
    with the same signature as pydantic's `BaseModel.model_dump` method.

    Usage example:
    ```python
        from pydantic import BaseModel

        from melting_schemas.utils import to_openapi_examples, wrap


        class NamedModel(BaseModel):
            name: str
            age: int


        # Named models can be used as is
        examples = to_openapi_examples([NamedModel(name="John", age=30)])
        print(examples)


        class UnnamedModel(BaseModel):
            age: int


        # Unnamed models need to be wrapped to be used
        example = wrap(name="John", value=UnnamedModel(age=30))
        examples = to_openapi_examples([example])
        print(examples)
    ```

    Args:
        examples: A list of NamedDumpable compatible objects or OpenAPIExample objects.

    Returns:
        A dictionary of OpenAPI Examples
    """

    openapi_examples = {}
    for example in examples:
        if isinstance(example, OpenAPIExample):
            value = example.value.model_dump(by_alias=True, exclude_unset=True)
        else:
            value = example.model_dump(by_alias=True, exclude_unset=True)
        openapi_examples[example.name] = Example(value=value)
    return openapi_examples



from typing import Annotated, Any, Literal

from pydantic import BaseModel, Field


class ChatMLMessage(BaseModel):
    content: str
    name: Annotated[str, Field(pattern=r"^[a-zA-Z0-9_]*$", max_length=64)] | None = None
    role: Literal["user", "assistant", "system"]


class ToolCallFunction(BaseModel):
    arguments: str
    name: str


class ToolCall(BaseModel):
    id: str
    type: Literal["function"]
    function: ToolCallFunction


class ToolCallMLMessage(BaseModel):
    id: str
    content: None
    tool_calls: list[ToolCall]
    role: Literal["assistant"]


class ToolMLMessage(BaseModel):
    tool_id: str
    content: str
    name: str
    role: Literal["tool"]


BufferedMLMessageType = ChatMLMessage | ToolCallMLMessage | ToolMLMessage



from prompts.turbo import PromptRole, TemplateInputs

from melting_schemas.utils import wrap

from ..completion.chat import ChatModelSettings
from .prompt import ChatPromptTemplate, Template


def chat_prompt_template_examples():
    minimal_prompt_template = ChatPromptTemplate(
        assistant_templates="<text>",
        description="Single of its kind, example app, teia org.",
        name="teia.example_app.single.example01",
        system_templates="<text>",
        user_templates="<text>",
        initial_template_data=[
            TemplateInputs(role=PromptRole.SYSTEM, name="", inputs={"a": "a"}, template_name="name")
        ],
        settings=ChatModelSettings(model="gpt-3.5-turbo"),
    )

    timeaware_prompt_template = ChatPromptTemplate(
        assistant_templates="<text>",
        description="Single of its kind, example app, teia org.",
        name="teia.example.1",
        system_templates="Current timestamp: <now>\nYou are a helpful chatbot.",
        user_templates="<text>",
        settings=ChatModelSettings(model="gpt-3.5-turbo"),
    )

    many_template = ChatPromptTemplate(
        name="teia.example.2",
        description="A development example.",
        system_templates=[
            Template(template_name="plugin_prompt", template="<plugin_data>"),
        ],
        user_templates=[
            Template(template_name="user_prompt", template="<question>"),
        ],
        assistant_templates=[
            Template(template_name="assistant_prompt", template="<message>"),
        ],
        settings=ChatModelSettings(
            model="gpt-3.5-turbo",
            max_tokens=200,
            temperature=0.25,
        ),
    )

    return [
        wrap(name="Minimal prompt template", value=minimal_prompt_template),
        wrap(name="Time-aware prompt template", value=timeaware_prompt_template),
        wrap(name="Many template", value=many_template),
    ]



from .examples import chat_prompt_template_examples
from .prompt import (
    ChatPrompt,
    ChatPromptTemplate,
    CompleteChatPromptTemplate,
    CompleteTextPromptTemplate,
    CreateCompletionPrompt,
    GeneratedFields,
    GetCompletionPrompt,
    UpdateChatPrompt,
    UpdateCompletionPrompt,
)



from pydantic import BaseModel

from .completion.finish_reason import FinishReason


class TokenUsage(BaseModel):
    prompt_tokens: int
    total_tokens: int
    completion_tokens: int | None = None


class Timings(BaseModel):
    total: float


class StreamTimings(BaseModel):
    avg: float
    first: float
    max: float
    min: float
    total: float


class StreamUsageInfo(BaseModel):
    finish_reason: FinishReason
    token_usage: TokenUsage
    timings: StreamTimings



from typing import Any, Literal

from pydantic import BaseModel, Field, SerializeAsAny

from melting_schemas.json_schema import FunctionJsonSchema

from .params import Forwardings, GeneratedParams, StaticParams


class NoopToolCallee(BaseModel):
    type: Literal["noop"] = "noop"


class HttpToolCallee(BaseModel):
    type: Literal["http"] = "http"
    method: Literal["GET", "POST"]
    headers: dict[str, str] = Field(default_factory=dict)
    url: str
    forwardings: Forwardings = Field(default_factory=Forwardings)
    static_params: StaticParams = Field(default_factory=StaticParams)
    generated_params: GeneratedParams = Field(default_factory=GeneratedParams)
    parameter_translator: dict[str, str] = Field(default_factory=dict)


class ToolJsonSchema(BaseModel):
    type: Literal["function"] = "function"
    function: FunctionJsonSchema


class ToolSpec(BaseModel):
    tool_name: str
    callee: HttpToolCallee | NoopToolCallee
    json_schema: ToolJsonSchema


class ToolInfo(BaseModel):
    spec: SerializeAsAny[ToolSpec | FunctionJsonSchema]
    ml_message_id: Any = None


class CallableStaticTool(BaseModel):
    name: str
    arguments: dict[str, Any]
    callee: HttpToolCallee | NoopToolCallee
    ml_message_id: Any = None



from melting_schemas.json_schema import FunctionJsonSchema, ObjectSchema, StringSchema
from melting_schemas.utils import wrap

from ..buffered_ml_messages import ChatMLMessage
from ..templating import TemplateInputs
from .params import Forwardings, StaticTool
from .requests import PromptedTCallRequest, RawTCallRequest
from .settings import TCallModelSettings
from .specs import (
    GeneratedParams,
    HttpToolCallee,
    StaticParams,
    ToolJsonSchema,
    ToolSpec,
)


def raw_tcall_request_examples():
    json_schema_tcall = RawTCallRequest(
        tools=[
            ToolJsonSchema(
                type="function",
                function=FunctionJsonSchema(
                    name="weather",
                    description="Retrieve weather information based on a city's name.",
                    parameters=ObjectSchema(
                        type="object",
                        properties={
                            "city_name": StringSchema(
                                type="string",
                                description="Name of the city.",
                            )
                        },
                        required=["city_name"],
                    ),
                ),
            )
        ],
        messages=[ChatMLMessage(content="How is the weather in Paris?", role="user")],
        settings=TCallModelSettings(model="gpt-4o"),
    )

    tool_spec_tcall = RawTCallRequest(
        tools=[
            ToolSpec(
                tool_name="weather_tool",
                callee=HttpToolCallee(
                    method="GET",
                    forwardings=Forwardings(headers=["authorization"]),
                    url="https://plugins.beta.allai.digital/weather-plugin/weather_current",
                    static_params=StaticParams(query={"kwargs": ""}),
                    generated_params=GeneratedParams(query=["place"]),
                    parameter_translator={"city_name": "place"},
                ),
                json_schema=ToolJsonSchema(
                    type="function",
                    function=FunctionJsonSchema(
                        name="weather",
                        description="Retrieve weather information based on a city's name.",
                        parameters=ObjectSchema(
                            type="object",
                            properties={
                                "city_name": StringSchema(
                                    type="string",
                                    description="Name of the city.",
                                )
                            },
                            required=["city_name"],
                        ),
                    ),
                ),
            )
        ],
        messages=[ChatMLMessage(content="How is the weather in Paris?", role="user")],
        settings=TCallModelSettings(model="gpt-4o"),
    )

    native_tcall = RawTCallRequest(
        tools=["weather_plugin"],
        messages=[ChatMLMessage(content="How is the weather in Paris?", role="user")],
        settings=TCallModelSettings(model="gpt-4o"),
    )

    return [
        wrap(name="Raw Json Schema Request", value=json_schema_tcall),
        wrap(name="Tool Calling Request", value=tool_spec_tcall),
        wrap(name="Native Tool Calling Request", value=native_tcall),
    ]


def prompted_tcall_request_examples():
    json_schema_tcall = PromptedTCallRequest(
        tools=[
            ToolJsonSchema(
                type="function",
                function=FunctionJsonSchema(
                    name="weather",
                    description="Retrieve weather information based on a city's name.",
                    parameters=ObjectSchema(
                        type="object",
                        properties={
                            "city_name": StringSchema(
                                type="string",
                                description="Name of the city.",
                            )
                        },
                        required=["city_name"],
                    ),
                ),
            )
        ],
        prompt_name="weather_prompt",
        prompt_inputs=[TemplateInputs(inputs={"city_name": "Paris"}, role="user")],
        settings=TCallModelSettings(model="gpt-4o"),
    )

    tool_spec_tcall = PromptedTCallRequest(
        tools=[
            ToolSpec(
                tool_name="weather_tool",
                callee=HttpToolCallee(
                    method="GET",
                    forwardings=Forwardings(headers=["authorization"]),
                    url="https://plugins.beta.allai.digital/weather-plugin/weather_current",
                    static_params=StaticParams(query={"kwargs": ""}),
                    generated_params=GeneratedParams(query=["place"]),
                    parameter_translator={"city_name": "place"},
                ),
                json_schema=ToolJsonSchema(
                    type="function",
                    function=FunctionJsonSchema(
                        name="weather",
                        description="Retrieve weather information based on a city's name.",
                        parameters=ObjectSchema(
                            type="object",
                            properties={
                                "city_name": StringSchema(
                                    type="string",
                                    description="Name of the city.",
                                )
                            },
                            required=["city_name"],
                        ),
                    ),
                ),
            )
        ],
        prompt_name="weather_prompt",
        prompt_inputs=[TemplateInputs(inputs={"city_name": "Paris"}, role="user")],
        settings=TCallModelSettings(model="gpt-4o"),
    )

    native_tcall = PromptedTCallRequest(
        tools=["weather_plugin"],
        prompt_name="weather_prompt",
        prompt_inputs=[TemplateInputs(inputs={"city_name": "Paris"}, role="user")],
        settings=TCallModelSettings(model="gpt-4o"),
    )

    return [
        wrap(name="Raw Prompted Json Schema Request", value=json_schema_tcall),
        wrap(name="Tool Calling Prompted Request", value=tool_spec_tcall),
        wrap(name="Native Tool Calling Prompted Request", value=native_tcall),
    ]


def special_tcall_request_examples():
    json_schema_tcall = RawTCallRequest(
        tools=[
            ToolJsonSchema(
                type="function",
                function=FunctionJsonSchema(
                    name="weather",
                    description="Retrieve weather information based on a city's name.",
                    parameters=ObjectSchema(
                        type="object",
                        properties={
                            "city_name": StringSchema(
                                type="string",
                                description="Name of the city.",
                            )
                        },
                        required=["city_name"],
                    ),
                ),
            )
        ],
        static_tools=[StaticTool(name="locator", arguments={}, response="User lives in Paris")],
        messages=[ChatMLMessage(content="How is the weather in Paris?", role="user")],
        settings=TCallModelSettings(model="gpt-4o"),
    )

    tool_spec_tcall = RawTCallRequest(
        tools=[
            ToolSpec(
                tool_name="weather_tool",
                callee=HttpToolCallee(
                    method="GET",
                    forwardings=Forwardings(headers=["authorization"]),
                    url="https://plugins.beta.allai.digital/weather-plugin/weather_current",
                    static_params=StaticParams(query={"kwargs": ""}),
                    generated_params=GeneratedParams(query=["place"]),
                    parameter_translator={"city_name": "place"},
                ),
                json_schema=ToolJsonSchema(
                    type="function",
                    function=FunctionJsonSchema(
                        name="weather",
                        description="Retrieve weather information based on a city's name.",
                        parameters=ObjectSchema(
                            type="object",
                            properties={
                                "city_name": StringSchema(
                                    type="string",
                                    description="Name of the city.",
                                )
                            },
                            required=["city_name"],
                        ),
                    ),
                ),
            )
        ],
        static_tools=[StaticTool(name="locator", arguments={}, response="User lives in Paris")],
        messages=[ChatMLMessage(content="How is the weather in Paris?", role="user")],
        settings=TCallModelSettings(model="gpt-4o"),
    )

    native_tcall = RawTCallRequest(
        tools=["weather_plugin"],
        static_tools=[StaticTool(name="locator", arguments={}, response="User lives in Paris")],
        messages=[ChatMLMessage(content="How is the weather in Paris?", role="user")],
        settings=TCallModelSettings(model="gpt-4o"),
    )

    return [
        wrap(name="Raw Json Schema Request", value=json_schema_tcall),
        wrap(name="Tool Calling Request", value=tool_spec_tcall),
        wrap(name="Native Tool Calling Request", value=native_tcall),
    ]



from typing import Literal

from pydantic import BaseModel


class TemplateInputs(BaseModel):
    inputs: dict[str, str]
    name: str | None = None
    role: Literal["user", "system", "assistant"]
    template_name: str | None = None  # Advanced usage: select sub-templates


class Templating(BaseModel):
    prompt_inputs: list[TemplateInputs | dict]
    prompt_id: str
    prompt_name: str



from typing import Any, Literal

from pydantic import BaseModel


class ChatMLMessageChunk(BaseModel):
    id: Any
    delta: str


class ToolCallFunctionChunk(BaseModel):
    delta: str
    name: str | None = None


class ToolCallChunk(BaseModel):
    id: str
    tool_id: str
    type: Literal["function"]
    function: ToolCallFunctionChunk


class ToolMLMessageChunk(BaseModel):
    tool_id: str
    delta: str
    name: str
    role: Literal["tool"]



from melting_schemas.utils import wrap

from .text_encoding import RawTextEncodingRequest


def raw_text_encoding_examples():
    minimal = RawTextEncodingRequest(
        model="sentence-transformers/all-MiniLM-L6-v2",
        snippets=["I like to eat apples."],
    )
    return [wrap(name="Minimal", value=minimal)]



from ..buffered_ml_messages import (
    ChatMLMessage,
    ToolCall,
    ToolCallFunction,
    ToolCallMLMessage,
    ToolMLMessage,
)
from ..streamed_ml_messages import (
    ChatMLMessageChunk,
    ToolCallChunk,
)
from .params import GeneratedParams, StaticParams, StaticTool
from .requests import HybridTCallRequest, PromptedTCallRequest, RawTCallRequest
from .settings import TCallModelSettings
from .specs import (
    CallableStaticTool,
    HttpToolCallee,
    NoopToolCallee,
    ToolInfo,
    ToolJsonSchema,
    ToolSpec,
)



from ..buffered_ml_messages import ChatMLMessage
from ..streamed_ml_messages import ChatMLMessageChunk
from .examples import text_completion_examples
from .requests import RawTextCompletionRequest
from .settings import TextModelSettings
